
;// https://docs.microsoft.com/en-us/windows-hardware/drivers/gettingstarted/minidrivers-and-driver-pairs
;// Module Name: io.h
;//---------------------------------------

include 'ntstatus.inc'

;// *******************************
;//
;// Define the major function codes for IRPs.
;//

IRP_MJ_CREATE                   =  0x00
IRP_MJ_CREATE_NAMED_PIPE        =  0x01
IRP_MJ_CLOSE                    =  0x02
IRP_MJ_READ                     =  0x03
IRP_MJ_WRITE                    =  0x04
IRP_MJ_QUERY_INFORMATION        =  0x05
IRP_MJ_SET_INFORMATION          =  0x06
IRP_MJ_QUERY_EA                 =  0x07
IRP_MJ_SET_EA                   =  0x08
IRP_MJ_FLUSH_BUFFERS            =  0x09
IRP_MJ_QUERY_VOLUME_INFORMATION =  0x0a
IRP_MJ_SET_VOLUME_INFORMATION   =  0x0b
IRP_MJ_DIRECTORY_CONTROL        =  0x0c
IRP_MJ_FILE_SYSTEM_CONTROL      =  0x0d
IRP_MJ_DEVICE_CONTROL           =  0x0e
IRP_MJ_INTERNAL_DEVICE_CONTROL  =  0x0f
IRP_MJ_SHUTDOWN                 =  0x10
IRP_MJ_LOCK_CONTROL             =  0x11
IRP_MJ_CLEANUP                  =  0x12
IRP_MJ_CREATE_MAILSLOT          =  0x13
IRP_MJ_QUERY_SECURITY           =  0x14
IRP_MJ_SET_SECURITY             =  0x15
IRP_MJ_POWER                    =  0x16
IRP_MJ_SYSTEM_CONTROL           =  0x17
IRP_MJ_DEVICE_CHANGE            =  0x18
IRP_MJ_QUERY_QUOTA              =  0x19
IRP_MJ_SET_QUOTA                =  0x1a
IRP_MJ_PNP                      =  0x1b
IRP_MJ_MAXIMUM_FUNCTION         =  0x1b
IRP_MJ_SCSI                     =  IRP_MJ_INTERNAL_DEVICE_CONTROL

;// Define the minor function codes for IRPs.
;//   0x00  0x7f   Microsoft.
;//   0x80  0xff   .

;// Directory control minor function codes

IRP_MN_QUERY_DIRECTORY          =  0x01
IRP_MN_NOTIFY_CHANGE_DIRECTORY  =  0x02

;// File system control minor function codes.

IRP_MN_KERNEL_CALL              =  0x04
IRP_MN_LOAD_FILE_SYSTEM         =  0x03
IRP_MN_MOUNT_VOLUME             =  0x01
IRP_MN_TRACK_LINK               =  0x04    ;// To be obsoleted soon
IRP_MN_USER_FS_REQUEST          =  0x00
IRP_MN_VERIFY_VOLUME            =  0x02

;// Lock control minor function codes

IRP_MN_LOCK                     =  0x01
IRP_MN_UNLOCK_ALL               =  0x03
IRP_MN_UNLOCK_ALL_BY_KEY        =  0x04
IRP_MN_UNLOCK_SINGLE            =  0x02

;//     . 
;// ,      FO_NO_INTERMEDIATE_BUFFERING, 
;//     ( Irp    ). 
;// IRP_MN_NORMAL    io,    fs.

IRP_MN_COMPLETE                 =  0x04
IRP_MN_COMPLETE_MDL             =  0x06
IRP_MN_COMPLETE_MDL_DPC         =  0x07
IRP_MN_COMPRESSED               =  0x08
IRP_MN_DPC                      =  0x01
IRP_MN_MDL                      =  0x02
IRP_MN_MDL_DPC                  =  0x03
IRP_MN_NORMAL                   =  0x00

;// SCSI & PNP minor function codes.

IRP_MN_CANCEL_REMOVE_DEVICE         =  0x03
IRP_MN_CANCEL_STOP_DEVICE           =  0x06
IRP_MN_DEVICE_USAGE_NOTIFICATION    =  0x16
IRP_MN_EJECT                        =  0x11
IRP_MN_FILTER_RESOURCE_REQUIREMENTS =  0x0D
IRP_MN_QUERY_BUS_INFORMATION        =  0x15
IRP_MN_QUERY_CAPABILITIES           =  0x09
IRP_MN_QUERY_DEVICE_RELATIONS       =  0x07
IRP_MN_QUERY_DEVICE_TEXT            =  0x0C
IRP_MN_QUERY_ID                     =  0x13
IRP_MN_QUERY_INTERFACE              =  0x08
IRP_MN_QUERY_LEGACY_BUS_INFORMATION =  0x18
IRP_MN_QUERY_PNP_DEVICE_STATE       =  0x14
IRP_MN_QUERY_REMOVE_DEVICE          =  0x01
IRP_MN_QUERY_RESOURCE_REQUIREMENTS  =  0x0B
IRP_MN_QUERY_RESOURCES              =  0x0A
IRP_MN_QUERY_STOP_DEVICE            =  0x05
IRP_MN_READ_CONFIG                  =  0x0F
IRP_MN_REMOVE_DEVICE                =  0x02
IRP_MN_SCSI_CLASS                   =  0x01
IRP_MN_SET_LOCK                     =  0x12
IRP_MN_START_DEVICE                 =  0x00
IRP_MN_STOP_DEVICE                  =  0x04
IRP_MN_SURPRISE_REMOVAL             =  0x17
IRP_MN_WRITE_CONFIG                 =  0x10

;// POWER minor function codes

IRP_MN_POWER_SEQUENCE               =  0x01
IRP_MN_QUERY_POWER                  =  0x03
IRP_MN_SET_POWER                    =  0x02
IRP_MN_WAIT_WAKE                    =  0x00

;// WMI minor function codes under IRP_MJ_SYSTEM_CONTROL

IRP_MN_CHANGE_SINGLE_INSTANCE       =  0x02
IRP_MN_CHANGE_SINGLE_ITEM           =  0x03
IRP_MN_DISABLE_COLLECTION           =  0x07
IRP_MN_DISABLE_EVENTS               =  0x05
IRP_MN_ENABLE_COLLECTION            =  0x06
IRP_MN_ENABLE_EVENTS                =  0x04
IRP_MN_EXECUTE_METHOD               =  0x09
IRP_MN_QUERY_ALL_DATA               =  0x00
IRP_MN_QUERY_SINGLE_INSTANCE        =  0x01
IRP_MN_REGINFO                      =  0x08
IRP_MN_REGINFO_EX                   =  0x0b
;// WMI minor function
IRP_MN_SET_TRACE_NOTIFY             =  0x0A

;// **********************************************

;//   Type -  

IO_TYPE_ADAPTER                 =  0x00000001
IO_TYPE_CONTROLLER              =  0x00000002
IO_TYPE_DEVICE                  =  0x00000003
IO_TYPE_DEVICE_OBJECT_EXTENSION =  0x0000000d
IO_TYPE_DRIVER                  =  0x00000004
IO_TYPE_ERROR_LOG               =  0x0000000b
IO_TYPE_ERROR_MESSAGE           =  0x0000000c
IO_TYPE_FILE                    =  0x00000005
IO_TYPE_IRP                     =  0x00000006
IO_TYPE_MASTER_ADAPTER          =  0x00000007
IO_TYPE_OPEN_PACKET             =  0x00000008
IO_TYPE_TIMER                   =  0x00000009
IO_TYPE_VPB                     =  0x0000000a

;// Define flags 

IO_ATTACH_DEVICE                =  0x00000400
IO_ATTACH_DEVICE_API            =  0x80000000
IO_CHECK_CREATE_PARAMETERS      =  0x00000200
IO_FORCE_ACCESS_CHECK           =  0x00000001
IO_IGNORE_SHARE_ACCESS_CHECK    =  0x00000800  ;// Ignores share access checks on opens.
IO_NO_PARAMETER_CHECKING        =  0x00000100
IO_OPEN_PAGING_FILE             =  0x00000002
IO_OPEN_TARGET_DIRECTORY        =  0x00000004
IO_REMOUNT                      =  0x00000001
IO_REPARSE                      =  0x00000000

;// Actions accepted by DRIVER_WRITE_PENDING

IO_DUMP_WRITE_DATA_PAGES  =  2
IO_DUMP_WRITE_FINISH      =  3   ;// finish pending IO
IO_DUMP_WRITE_FULFILL     =  0   ;// fulfill IO request as if DRIVER_WAIT
IO_DUMP_WRITE_INIT        =  4   ;// initialize locals
IO_DUMP_WRITE_RESUME      =  2   ;// resume pending IO
IO_DUMP_WRITE_START       =  1   ;// start new IO
; IO_DUMP_WRITE_DATA_SIZE   =  IO_DUMP_WRITE_DATA_PAGES << PAGE_SHIFT

;// Device Object (DO) flags

DO_BUFFERED_IO                  =  0x00000004
DO_BUS_ENUMERATED_DEVICE        =  0x00001000
DO_DEVICE_HAS_NAME              =  0x00000040
DO_DEVICE_INITIALIZING          =  0x00000080
DO_DIRECT_IO                    =  0x00000010
DO_EXCLUSIVE                    =  0x00000008
DO_LONG_TERM_REQUESTS           =  0x00000200
DO_LOW_PRIORITY_FILESYSTEM      =  0x00010000
DO_MAP_IO_BUFFER                =  0x00000020
DO_NEVER_LAST_DEVICE            =  0x00000400
DO_POWER_INRUSH                 =  0x00004000
DO_POWER_NOOP                   =  0x00008000
DO_POWER_PAGABLE                =  0x00002000
DO_SHUTDOWN_REGISTERED          =  0x00000800
DO_SYSTEM_BOOT_PARTITION        =  0x00000100
DO_VERIFY_VOLUME                =  0x00000002
DO_XIP                          =  0x00020000

;// Define the Device Object Extension Flags

DOE_DELETE_PENDING              =  0x00000002
DOE_REMOVE_PENDING              =  0x00000004
DOE_REMOVE_PROCESSED            =  0x00000008
DOE_START_PENDING               =  0x00000010
DOE_STARTIO_CANCELABLE          =  0x00000080
DOE_STARTIO_DEFERRED            =  0x00000100  ;// Use non-recursive startio
DOE_STARTIO_NO_CANCEL           =  0x00000200  ;// Pass non-cancelable IRP to startio
DOE_STARTIO_REQUESTED           =  0x00000020
DOE_STARTIO_REQUESTED_BYKEY     =  0x00000040
DOE_UNLOAD_PENDING              =  0x00000001

;// Define the Device Object characteristics flags

FILE_AUTOGENERATED_DEVICE_NAME           =  0x00000080
FILE_CHARACTERISTIC_CSV                  =  0x00010000
FILE_CHARACTERISTIC_PNP_DEVICE           =  0x00000800
FILE_CHARACTERISTIC_TS_DEVICE            =  0x00001000
FILE_CHARACTERISTIC_WEBDAV_DEVICE        =  0x00002000
FILE_DEVICE_ALLOW_APPCONTAINER_TRAVERSAL =  0x00020000
FILE_DEVICE_IS_MOUNTED                   =  0x00000020
FILE_DEVICE_SECURE_OPEN                  =  0x00000100
FILE_FLOPPY_DISKETTE                     =  0x00000004
FILE_PORTABLE_DEVICE                     =  0x00040000
FILE_READ_ONLY_DEVICE                    =  0x00000002
FILE_REMOTE_DEVICE                       =  0x00000010
FILE_REMOVABLE_MEDIA                     =  0x00000001
FILE_VIRTUAL_VOLUME                      =  0x00000040
FILE_WRITE_ONCE_MEDIA                    =  0x00000008

;// Define Driver Object (DRVO) flags

DRVO_BASE_FILESYSTEM_DRIVER     =  0x00000080   ;// A driver that is at the bottom of the filesystem stack.
DRVO_BOOTREINIT_REGISTERED      =  0x00000020
DRVO_BUILTIN_DRIVER             =  0x00000004    ;// Driver objects for Hal, PnP Mgr
DRVO_INITIALIZED                =  0x00000010
DRVO_LEGACY_DRIVER              =  0x00000002
DRVO_LEGACY_RESOURCES           =  0x00000040
DRVO_REINIT_REGISTERED          =  0x00000008
DRVO_UNLOAD_INVOKED             =  0x00000001

;// Define File Object (FO) flags

FO_ALERTABLE_IO                 =  0x00000004
FO_CACHE_SUPPORTED              =  0x00000040
FO_CLEANUP_COMPLETE             =  0x00004000
FO_DELETE_ON_CLOSE              =  0x00010000
FO_DIRECT_DEVICE_OPEN           =  0x00000800
FO_FILE_FAST_IO_READ            =  0x00080000
FO_FILE_MODIFIED                =  0x00001000
FO_FILE_OBJECT_HAS_EXTENSION    =  0x00800000
FO_FILE_OPEN                    =  0x00000001
FO_FILE_OPEN_CANCELLED          =  0x00200000
FO_FILE_SIZE_CHANGED            =  0x00002000
FO_GENERATE_AUDIT_ON_CLOSE      =  0x00000400
FO_HANDLE_CREATED               =  0x00040000
FO_MAILSLOT                     =  0x00000200
FO_NAMED_PIPE                   =  0x00000080
FO_NO_INTERMEDIATE_BUFFERING    =  0x00000008
FO_OPENED_CASE_SENSITIVE        =  0x00020000
FO_RANDOM_ACCESS                =  0x00100000
FO_REMOTE_ORIGIN                =  0x01000000
FO_SEQUENTIAL_ONLY              =  0x00000020
FO_STREAM_FILE                  =  0x00000100
FO_SYNCHRONOUS_IO               =  0x00000002
FO_TEMPORARY_FILE               =  0x00008000
FO_VOLUME_OPEN                  =  0x00400000
FO_WRITE_THROUGH                =  0x00000010

;// Define I/O Request Packet (IRP) flags

IRP_ASSOCIATED_IRP              =  0x00000008
IRP_BUFFERED_IO                 =  0x00000010
IRP_CLOSE_OPERATION             =  0x00000400
IRP_CREATE_OPERATION            =  0x00000080
IRP_DEALLOCATE_BUFFER           =  0x00000020
IRP_DEFER_IO_COMPLETION         =  0x00000800
IRP_HIGH_PRIORITY_PAGING_IO     =  0x00008000
IRP_HOLD_DEVICE_QUEUE           =  0x00002000
IRP_INPUT_OPERATION             =  0x00000040
IRP_MOUNT_COMPLETION            =  0x00000002
IRP_NOCACHE                     =  0x00000001
IRP_OB_QUERY_NAME               =  0x00001000
IRP_PAGING_IO                   =  0x00000002
IRP_READ_OPERATION              =  0x00000100
IRP_RETRY_IO_COMPLETION         =  0x00004000
IRP_SYNCHRONOUS_API             =  0x00000004
IRP_SYNCHRONOUS_PAGING_IO       =  0x00000040
IRP_VERIFIER_MASK               =  0xC0000000
IRP_WRITE_OPERATION             =  0x00000200
IRP_SET_USER_EVENT              =  IRP_CLOSE_OPERATION

;// Define I/O request packet (IRP) alternate flags for allocation control.

IRP_ALLOCATED_FIXED_SIZE        =  0x04
IRP_ALLOCATED_MUST_SUCCEED      =  0x02
IRP_LOOKASIDE_ALLOCATION        =  0x08
IRP_QUOTA_CHARGED               =  0x01

;// Define stack location control flags

SL_ERROR_RETURNED               =  0x02
SL_INVOKE_ON_CANCEL             =  0x20
SL_INVOKE_ON_ERROR              =  0x80
SL_INVOKE_ON_SUCCESS            =  0x40
SL_PENDING_RETURNED             =  0x01

;// Define flags for various functions

SL_ALLOW_RAW_MOUNT              =  0x01
SL_CASE_SENSITIVE               =  0x80
SL_EXCLUSIVE_LOCK               =  0x02
SL_FAIL_IMMEDIATELY             =  0x01
SL_FORCE_ACCESS_CHECK           =  0x01
SL_FT_SEQUENTIAL_WRITE          =  0x08
SL_INDEX_SPECIFIED              =  0x04
SL_KEY_SPECIFIED                =  0x01
SL_OPEN_PAGING_FILE             =  0x02
SL_OPEN_TARGET_DIRECTORY        =  0x04
SL_OVERRIDE_VERIFY_VOLUME       =  0x02
SL_READ_ACCESS_GRANTED          =  0x01
SL_RESTART_SCAN                 =  0x01
SL_RETURN_SINGLE_ENTRY          =  0x02
SL_WATCH_TREE                   =  0x01
SL_WRITE_ACCESS_GRANTED         =  0x04
SL_WRITE_THROUGH                =  0x04

;//  DeviceType   DEVICE_OBJECT 
;//   IoCreateDevice()

FILE_DEVICE_8042_PORT           =  0x00000027
FILE_DEVICE_ACPI                =  0x00000032
FILE_DEVICE_BATTERY             =  0x00000029
FILE_DEVICE_BEEP                =  0x00000001
FILE_DEVICE_BUS_EXTENDER        =  0x0000002a
FILE_DEVICE_CD_ROM              =  0x00000002
FILE_DEVICE_CD_ROM_FILE_SYSTEM  =  0x00000003
FILE_DEVICE_CHANGER             =  0x00000030
FILE_DEVICE_CONTROLLER          =  0x00000004
FILE_DEVICE_DATALINK            =  0x00000005
FILE_DEVICE_DFS                 =  0x00000006
FILE_DEVICE_DFS_FILE_SYSTEM     =  0x00000035
FILE_DEVICE_DFS_VOLUME          =  0x00000036
FILE_DEVICE_DISK                =  0x00000007
FILE_DEVICE_DISK_FILE_SYSTEM    =  0x00000008
FILE_DEVICE_DVD                 =  0x00000033
FILE_DEVICE_FILE_SYSTEM         =  0x00000009
FILE_DEVICE_FIPS                =  0x0000003a
FILE_DEVICE_FULLSCREEN_VIDEO    =  0x00000034
FILE_DEVICE_INPORT_PORT         =  0x0000000a
FILE_DEVICE_KEYBOARD            =  0x0000000b
FILE_DEVICE_KS                  =  0x0000002f
FILE_DEVICE_KSEC                =  0x00000039
FILE_DEVICE_MAILSLOT            =  0x0000000c
FILE_DEVICE_MASS_STORAGE        =  0x0000002d
FILE_DEVICE_MIDI_IN             =  0x0000000d
FILE_DEVICE_MIDI_OUT            =  0x0000000e
FILE_DEVICE_MODEM               =  0x0000002b
FILE_DEVICE_MOUSE               =  0x0000000f
FILE_DEVICE_MULTI_UNC_PROVIDER  =  0x00000010
FILE_DEVICE_NAMED_PIPE          =  0x00000011
FILE_DEVICE_NETWORK             =  0x00000012
FILE_DEVICE_NETWORK_BROWSER     =  0x00000013
FILE_DEVICE_NETWORK_FILE_SYSTEM =  0x00000014
FILE_DEVICE_NETWORK_REDIRECTOR  =  0x00000028
FILE_DEVICE_NULL                =  0x00000015
FILE_DEVICE_PARALLEL_PORT       =  0x00000016
FILE_DEVICE_PHYSICAL_NETCARD    =  0x00000017
FILE_DEVICE_PRINTER             =  0x00000018
FILE_DEVICE_SCANNER             =  0x00000019
FILE_DEVICE_SCREEN              =  0x0000001c
FILE_DEVICE_SERENUM             =  0x00000037
FILE_DEVICE_SERIAL_MOUSE_PORT   =  0x0000001a
FILE_DEVICE_SERIAL_PORT         =  0x0000001b
FILE_DEVICE_SMARTCARD           =  0x00000031
FILE_DEVICE_SMB                 =  0x0000002e
FILE_DEVICE_SOUND               =  0x0000001d
FILE_DEVICE_STREAMS             =  0x0000001e
FILE_DEVICE_TAPE                =  0x0000001f
FILE_DEVICE_TAPE_FILE_SYSTEM    =  0x00000020
FILE_DEVICE_TERMSRV             =  0x00000038
FILE_DEVICE_TRANSPORT           =  0x00000021
FILE_DEVICE_UNKNOWN             =  0x00000022
FILE_DEVICE_VDM                 =  0x0000002c
FILE_DEVICE_VIDEO               =  0x00000023
FILE_DEVICE_VIRTUAL_DISK        =  0x00000024
FILE_DEVICE_WAVE_IN             =  0x00000025
FILE_DEVICE_WAVE_OUT            =  0x00000026

;// IoCompleteRequest() priority increment

IO_CD_ROM_INCREMENT      =  1
IO_DISK_INCREMENT        =  1
IO_KEYBOARD_INCREMENT    =  6
IO_MAILSLOT_INCREMENT    =  2
IO_MOUSE_INCREMENT       =  6
IO_NAMED_PIPE_INCREMENT  =  2
IO_NETWORK_INCREMENT     =  2
IO_NO_INCREMENT          =  0
IO_PARALLEL_INCREMENT    =  1
IO_SERIAL_INCREMENT      =  2
IO_SOUND_INCREMENT       =  8
IO_VIDEO_INCREMENT       =  1

EVENT_INCREMENT          =  1
SEMAPHORE_INCREMENT      =  1

;// Define interface type

Cbus               =  9
Eisa               =  2
Internal           =  0
InternalPowerBus   =  13
Isa                =  1
MicroChannel       =  3
MPIBus             =  10
MPSABus            =  11
NuBus              =  7
PCIBus             =  5
PCMCIABus          =  8
PNPBus             =  15
PNPISABus          =  14
ProcessorInternal  =  12
TurboChannel       =  4
Vmcs               =  16
VMEBus             =  6

;// Define irq priority

IrqPriorityLow     =  1
IrqPriorityNormal  =  2
IrqPriorityHigh    =  3

;//*******************************************

struct IO_STACK_LOCATION
   MajorFunction     db  0
   MinorFunction     db  0
   Flags             db  0
   Control           db  0
   Parameters        dd  4 dup(0)
   DeviceObject      dd  0
   FileObject        dd  0
   CompletionRoutine dd  0
   Context           dd  0 
ends 

struct UNICODE_STRING
   Length           dw  0
   MaximumLength    dw  0
   Buffer           dd  0
ends

struct LIST_ENTRY
   Flink      dd  0
   Blink      dd  0 
ends

struct KDPC
   Type             db  0
   Importance       db  0
   Number           dw  0
   DpcListEntry     LIST_ENTRY
   DeferredRoutine  dd  0 
   DeferredContext  dd  0 
   SystemArgument1  dd  0 
   SystemArgument2  dd  0 
   DpcData          dd  0 
ends

struct DRIVER_EXTENSION
   DriverObject          dd  0  
   AddDevice             dd  0  
   Count                 dd  0  
   ServiceKeyName        UNICODE_STRING
   ClientDriverExtension dd  0    ; IO_CLIENT_EXTENSION
   FsFilterCallbacks     dd  0    ; FS_FILTER_CALLBACKS
ends   

struct NT_TIB
   ExceptionList         dd  0    ; EXCEPTION_REGISTRATION_RECORD
   StackBase             dd  0    ; 
   StackLimit            dd  0    ; 
   SubSystemTib          dd  0    ; 
   Version               dd  0    ; 
   ArbitraryUserPointer  dd  0    ; 
   Self                  dd  0    ; NT_TIB
ends
;//--------------------------------
;//    [FS:00]
;//--------------------------------
struct KPCR
   NtTib             NT_TIB
   SelfPcr           dd  0    ; KPCR
   Prcb              dd  0    ; KPRCB
   Irql              dd  0    ; 
   IRR               dd  0    ; 
   IrrActive         dd  0    ; 
   IDR               dd  0    ; 
   KdVersionBlock    dd  0    ; 
   IDT               dd  0    ; KIDTENTRY
   GDT               dd  0    ; KGDTENTRY
   TSS               dd  0    ; KTSS
   MajorVersion      dw  0    ; 
   MinorVersion      dw  0    ; 
   SetMember         dd  0    ; 
   StallScaleFactor  dd  0    ; 
   DebugActive       db  0    ; 
ends   
;//----------------------------

struct DRIVER_OBJECT
   Type                  dw  0 
   Size                  dw  0
   DeviceObject          dd  0
   Flags                 dd  0 
   DriverStart           dd  0
   DriverSize            dd  0
   DriverSection         dd  0     ;<-----// LDR_DATA_TABLE_ENTRY
   DriverExtension       dd  0
   DriverName            UNICODE_STRING
   HardwareDatabase      dd  0
   FastIoDispatch        dd  0
   DriverInit            dd  0
   DriverStartIo         dd  0
   DriverUnload          dd  0
;   MajorFunction         dd  28 dup(0)
   IRP_MJ_CREATE                   dd  0   ; 0x00
   IRP_MJ_CREATE_NAMED_PIPE        dd  0   ; 0x01
   IRP_MJ_CLOSE                    dd  0   ; 0x02
   IRP_MJ_READ                     dd  0   ; 0x03
   IRP_MJ_WRITE                    dd  0   ; 0x04
   IRP_MJ_QUERY_INFORMATION        dd  0   ; 0x05
   IRP_MJ_SET_INFORMATION          dd  0   ; 0x06
   IRP_MJ_QUERY_EA                 dd  0   ; 0x07
   IRP_MJ_SET_EA                   dd  0   ; 0x08
   IRP_MJ_FLUSH_BUFFERS            dd  0   ; 0x09
   IRP_MJ_QUERY_VOLUME_INFORMATION dd  0   ; 0x0a
   IRP_MJ_SET_VOLUME_INFORMATION   dd  0   ; 0x0b
   IRP_MJ_DIRECTORY_CONTROL        dd  0   ; 0x0c
   IRP_MJ_FILE_SYSTEM_CONTROL      dd  0   ; 0x0d
   IRP_MJ_DEVICE_CONTROL           dd  0   ; 0x0e
   IRP_MJ_INTERNAL_DEVICE_CONTROL  dd  0   ; 0x0f
   IRP_MJ_SHUTDOWN                 dd  0   ; 0x10
   IRP_MJ_LOCK_CONTROL             dd  0   ; 0x11
   IRP_MJ_CLEANUP                  dd  0   ; 0x12
   IRP_MJ_CREATE_MAILSLOT          dd  0   ; 0x13
   IRP_MJ_QUERY_SECURITY           dd  0   ; 0x14
   IRP_MJ_SET_SECURITY             dd  0   ; 0x15
   IRP_MJ_POWER                    dd  0   ; 0x16
   IRP_MJ_SYSTEM_CONTROL           dd  0   ; 0x17
   IRP_MJ_DEVICE_CHANGE            dd  0   ; 0x18
   IRP_MJ_QUERY_QUOTA              dd  0   ; 0x19
   IRP_MJ_SET_QUOTA                dd  0   ; 0x1a
   IRP_MJ_PNP                      dd  0   ; 0x1b
ends

struct DEVICE_OBJECT
   Type                  dw  0
   Size                  dw  0
   ReferenceCount        dd  0
   DriverObject          dd  0   ; DRIVER_OBJECT
   NextDevice            dd  0   ; DEVICE_OBJECT
   AttachedDevice        dd  0   ; DEVICE_OBJECT
   CurrentIrp            dd  0   ; IRP
   Timer                 dd  0   ; IO_TIMER
   Flags                 dd  0
   Characteristics       dd  0
   Vpb                   dd  0
   DeviceExtension       dd  0
   DeviceType            dd  0
   StackSize             dd  0
   Queue                 LIST_ENTRY
   AlignRequirement      dd  0
   DeviceQueue           db  20 dup(0)
   Dpc                   KDPC
   ActiveThreadCount     dd  0
   SecurityDescriptor    dd  0
   DeviceLock            db  16 dup(0)  ; KEVENT
   SectorSize            dd  0
   DeviceObjectExtension dd  0
   Reserved              dd  0
ends 

struct FILE_OBJECT
   Type               dw  0
   Size               dw  0
   Padding            dd  0
   DeviceObject       dd  0   ; _DEVICE_OBJECT
   Vpb                dd  0   ; _VPB
   FsContext          dd  0 
   FsContext2         dd  0 
   SectionObjPointer  dd  0   ; _SECTION_OBJECT_POINTERS
   PrivateCacheMap    dd  0 
   FinalStatus        dd  0
   RelatedFileObject  dd  0   ; _FILE_OBJECT
   LockOperation      db  0
   DeletePending      db  0
   ReadAccess         db  0
   WriteAccess        db  0
   DeleteAccess       db  0
   SharedRead         db  0
   SharedWrite        db  0
   SharedDelete       db  0
   Flags              dd  0
   FileName           UNICODE_STRING
   CurrentByteOffset  dd  0 
   Waiters            dd  0
   Busy               dd  0
   LastLock           dd  0 
   Lock               db  16 dup(0)   ; KEVENT
   Event              db  16 dup(0)   ; KEVENT
   CompletionContext  dd  0           ; IO_COMPLETION_CONTEXT
   IrpListLock        dd  0
   IrpList            LIST_ENTRY
   FileObjectExt      dd  0 
ends

struct IO_STATUS_BLOCK
   Status            dd  0  ;//  STATUS_xx   IoCompleteRequest()
   Information       dd  0  ;//  -  ,  
ends 

struct _IRP
   Type                  dw  0
   Size                  dw  0
   MdlAddress            dd  0      ; MDL
   Flags                 dd  0
   AssociatedIrp         dd  0   ;<-----------//  DeviceIoControl()
   ThreadListEntry       LIST_ENTRY
   IoStatus              IO_STATUS_BLOCK
   RequestorMode         db  0
   PendingReturned       db  0   ;<-----------//
   StackCount            db  0
   CurrentLocation       db  0
   Cancel                db  0
   CancelIrql            db  0
   ApcEnvironment        db  0
   AllocationFlags       db  0
   UserIosb              dd  0        ; IO_STATUS_BLOCK
   UserEvent             dd  0
   Overlay               dq  0
   CancelRoutine         dd  0
   UserBuffer            dd  0
   DriverContext         dd  4 dup(0) ; 0x078 Tail.Overlay
   Thread                dd  0        ; ETHREAD
   AuxiliaryBuffer       dd  0
   ListEntry             LIST_ENTRY
   CurrentStackLocation  dd  0        ; IO_STACK_LOCATION  --+
   OriginalFileObject    dd  0        ; FILE_OBJECT          |
ends ;                                                       |
;//                                                          |
;// IO_STACK_LOCATION Parameters.XX   <----------------------+
;//
struct PARAM_CREATE
   SecurityContext    dd  0  ; IO_SECURITY_CONTEXT
   Options            dd  0
   FileAttributes     dw  0
   ShareAccess        dw  0
   EaLength           dd  0
ends   
struct PARAM_READ_WRITE          ; Buffered = Irp->AssociatedIrp.SystemBuffer, Direct = Irp->MdlAddress
   Length             dd  0      ;    
   Key                dd  0
   ByteOffset         dd  0      ;   
   Reserved           dd  0
ends
struct PARAM_READ_WRITE_CONFIG
   WhichSpace         dd  0
   Buffer             dd  0
   Offset             dd  0
   Length             dd  0
ends
struct PARAM_IO_CONTROL          ;  PARAM_READ_WRITE ^^^
   OutputBuffLength   dd  0
   InputBuffLength    dd  0
   IoControlCode      dd  0
   Type3InputBuff     dd  0
ends

;//------------------------------------

struct FAST_MUTEX
   Count             dd  0
   Owner             dd  0    ; KTHREAD
   Contention        dd  0
   Event             dq  0,0  ; KEVENT
   OldIrql           dd  0
ends

struct DISPATCHER_HEADER
   Type             db  0
   Absolute         db  0
   Size             db  0
   Inserted         db  0
   SignalState      dd  0
   WaitListHead     LIST_ENTRY
ends
   
struct KPROCESS
   Header                DISPATCHER_HEADER
   ProfileListHead       LIST_ENTRY
   DirectoryTableBase    dd  0,0
   LdtDescriptor         dd  0,0   ; KGDTENTRY
   Int21Descriptor       dd  0,0   ; KIDTENTRY
   IopmOffset            dw  0
   Iopl                  db  0
   Unused                db  0
   ActiveProcessors      dd  0
   KernelTime            dd  0
   UserTime              dd  0
   ReadyListHead         LIST_ENTRY
   SwapListEntry         dd  0
   VdmTrapcHandler       dd  0
   ThreadListHead        LIST_ENTRY
   ProcessLock           dd  0
   Affinity              dd  0
   StackCount            dw  0
   BasePriority          db  0
   ThreadQuantum         db  0
   AutoAlignment         db  0
   State                 db  0
   ThreadSeed            db  0
   DisableBoost          db  0
   PowerState            db  0
   DisableQuantum        db  0
   IdealNode             db  0
   ExecuteOptions        db  0
ends 

struct EPROCESS
   Pcb                   KPROCESS
   ProcessLock           dd  0
   CreateTime            dq  0
   ExitTime              dq  0 
   RundownProtect        dd  0
   UniqueProcessId       dd  0
   ActiveProcessLinks    LIST_ENTRY     ; offset 0x88
   QuotaUsage            dd  0,0,0
   QuotaPeak             dd  0,0,0
   CommitCharge          dd  0
   PeakVirtualSize       dd  0
   VirtualSize           dd  0
   SessionProcessLinks   LIST_ENTRY
   DebugPort             dd  0
   ExceptionPort         dd  0
   ObjectTable           dd  0
   Token                 dd  0           ;  [2:0] = RefCount ()
   WorkingSetLock        FAST_MUTEX
   WorkingSetPage        dd  0
   AddressCreationLock   FAST_MUTEX
   HyperSpaceLock        dd  0
   ForkInProgress        dd  0           ; ETHREAD
   HardwareTrigger       dd  0
   VadRoot               dd  0
   VadHint               dd  0
   CloneRoot             dd  0
   NumberOfPrivatePages  dd  0
   NumberOfLockedPages   dd  0
   Win32Process          dd  0
   Job                   dd  0
   SectionObject         dd  0
   SectionBaseAddress    dd  0
   QuotaBlock            dd  0
   WorkingSetWatch       dd  0
   Win32WindowStation    dd  0
   InheritedFromPid      dd  0
   LdtInformation        dd  0
   VadFreeHint           dd  0
   VdmObjects            dd  0
   DeviceMap             dd  0
   PhysicalVadList       LIST_ENTRY
   PageDirectoryPte      dq  0          ; HARDWARE_PTE
   Session               dd  0
   ImageFileName         db  16 dup(0)
   JobLinks              LIST_ENTRY
   LockedPagesList       dd  0
   ThreadListHead        LIST_ENTRY
   SecurityPort          dd  0
   PaeTop                dd  0
   ActiveThreads         dd  0
   GrantedAccess         dd  0
   DefaultHardError      dd  0
   LastThreadExitStatus  dd  0
   Peb                   dd  0
   PrefetchTrace         dd  0      ; EX_FAST_REF
   ReadOperationCount    dq  0      ; LARGE_INTEGER
   WriteOperationCount   dq  0
   OtherOperationCount   dq  0
   ReadTransferCount     dq  0
   WriteTransferCount    dq  0
   OtherTransferCount    dq  0
   CommitChargeLimit     dd  0
   CommitChargePeak      dd  0
   AweInfo               dd  0
   SeAuditProcCreateInfo dd  0  ; SE_AUDIT_PROCESS_CREATION_INFO
ends

struct LDR_DATA_TABLE_ENTRY
   InLoadOrderLinks       LIST_ENTRY
   InMemoryOrderLinks     LIST_ENTRY
   InInitOrderLinks       LIST_ENTRY
   DllBase                dd  0
   EntryPoint             dd  0
   SizeOfImage            dd  0
   FullDllName            UNICODE_STRING
   BaseDllName            UNICODE_STRING
   Flags                  dd  0
   LoadCount              dw  0
   TlsIndex               dw  0
   HashLinks              LIST_ENTRY
   SectionPointer         dd  0
   CheckSum               dd  0
   TimeDateStamp          dd  0
   LoadedImports          dd  0
   EntryPointActContext   dd  0
   PatchInformation       dd  0
ends   

;//*******************************************
;//***************  MACRO  *******************

macro CompleteReq Status, Info {
       mov     eax,[pIrp]
       mov     dword[eax + _IRP.IoStatus],Status
       mov     dword[eax + _IRP.IoStatus+4],Info
       invoke  IoCompleteRequest,[pIrp],IO_NO_INCREMENT }

;//*******************************************

struct MDL
   Next             dd  0  ; MDL
   Size             dw  0 
   MdlFlags         dw  0  
   Process          dd  0  ; EPROCESS
   MappedSystemVa   dd  0
   StartVa          dd  0
   ByteCount        dd  0
   ByteOffset       dd  0
ends

;//  MmProbeAndLockPages() -   MDL
;//     MmUnlockPages()

IoReadAccess       =  0
IoWriteAccess      =  1
IoModifyAccess     =  2

KernelMode         =  0
UserMode           =  1
MaximumMode        =  2




























