
;// Module Name: io.h
;//----------------------------------

include 'ntstatus.inc'

FALSE				     =	0
TRUE	      			=	1
NonPagedPool  = 0
PagedPool     = 1

;// IRQL definitions

PASSIVE_LEVEL			=	0
APC_LEVEL		    	=	1
DISPATCH_LEVEL		=	2
CLOCK_LEVEL		  	=	13
IPI_LEVEL			    =	14	
POWER_LEVEL			  =	14	
PROFILE_LEVEL			=	15
HIGH_LEVEL			   =	15

;// *******************************
;//
;// Define the major function codes for IRPs.
;//

IRP_MJ_CREATE                   =  0x00
IRP_MJ_CREATE_NAMED_PIPE        =  0x01
IRP_MJ_CLOSE                    =  0x02
IRP_MJ_READ                     =  0x03
IRP_MJ_WRITE                    =  0x04
IRP_MJ_QUERY_INFORMATION        =  0x05
IRP_MJ_SET_INFORMATION          =  0x06
IRP_MJ_QUERY_EA                 =  0x07
IRP_MJ_SET_EA                   =  0x08
IRP_MJ_FLUSH_BUFFERS            =  0x09
IRP_MJ_QUERY_VOLUME_INFORMATION =  0x0a
IRP_MJ_SET_VOLUME_INFORMATION   =  0x0b
IRP_MJ_DIRECTORY_CONTROL        =  0x0c
IRP_MJ_FILE_SYSTEM_CONTROL      =  0x0d
IRP_MJ_DEVICE_CONTROL           =  0x0e
IRP_MJ_INTERNAL_DEVICE_CONTROL  =  0x0f
IRP_MJ_SHUTDOWN                 =  0x10
IRP_MJ_LOCK_CONTROL             =  0x11
IRP_MJ_CLEANUP                  =  0x12
IRP_MJ_CREATE_MAILSLOT          =  0x13
IRP_MJ_QUERY_SECURITY           =  0x14
IRP_MJ_SET_SECURITY             =  0x15
IRP_MJ_POWER                    =  0x16
IRP_MJ_SYSTEM_CONTROL           =  0x17
IRP_MJ_DEVICE_CHANGE            =  0x18
IRP_MJ_QUERY_QUOTA              =  0x19
IRP_MJ_SET_QUOTA                =  0x1a
IRP_MJ_PNP                      =  0x1b
IRP_MJ_MAXIMUM_FUNCTION         =  0x1b
IRP_MJ_SCSI                     =  IRP_MJ_INTERNAL_DEVICE_CONTROL

;// Define the minor function codes for IRPs.
;// The lower 128 codes, from 0x00 to 0x7f are reserved to Microsoft.  
;// The upper 128 codes, from 0x80 to 0xff are reserved to customers of Microsoft.

;// Directory control minor function codes

IRP_MN_QUERY_DIRECTORY          =  0x01
IRP_MN_NOTIFY_CHANGE_DIRECTORY  =  0x02

;// File system control minor function codes.

IRP_MN_KERNEL_CALL              =  0x04
IRP_MN_LOAD_FILE_SYSTEM         =  0x03
IRP_MN_MOUNT_VOLUME             =  0x01
IRP_MN_TRACK_LINK               =  0x04    ;// To be obsoleted soon
IRP_MN_USER_FS_REQUEST          =  0x00
IRP_MN_VERIFY_VOLUME            =  0x02

;// Lock control minor function codes

IRP_MN_LOCK                     =  0x01
IRP_MN_UNLOCK_ALL               =  0x03
IRP_MN_UNLOCK_ALL_BY_KEY        =  0x04
IRP_MN_UNLOCK_SINGLE            =  0x02

;//     . 
;// ,      FO_NO_INTERMEDIATE_BUFFERING, 
;//     ( Irp    ). 
;// IRP_MN_NORMAL    io,    fs.

IRP_MN_COMPLETE                 =  0x04
IRP_MN_COMPLETE_MDL             =  0x06
IRP_MN_COMPLETE_MDL_DPC         =  0x07
IRP_MN_COMPRESSED               =  0x08
IRP_MN_DPC                      =  0x01
IRP_MN_MDL                      =  0x02
IRP_MN_MDL_DPC                  =  0x03
IRP_MN_NORMAL                   =  0x00

;// SCSI & PNP minor function codes.

IRP_MN_CANCEL_REMOVE_DEVICE         =  0x03
IRP_MN_CANCEL_STOP_DEVICE           =  0x06
IRP_MN_DEVICE_USAGE_NOTIFICATION    =  0x16
IRP_MN_EJECT                        =  0x11
IRP_MN_FILTER_RESOURCE_REQUIREMENTS =  0x0D
IRP_MN_QUERY_BUS_INFORMATION        =  0x15
IRP_MN_QUERY_CAPABILITIES           =  0x09
IRP_MN_QUERY_DEVICE_RELATIONS       =  0x07
IRP_MN_QUERY_DEVICE_TEXT            =  0x0C
IRP_MN_QUERY_ID                     =  0x13
IRP_MN_QUERY_INTERFACE              =  0x08
IRP_MN_QUERY_LEGACY_BUS_INFORMATION =  0x18
IRP_MN_QUERY_PNP_DEVICE_STATE       =  0x14
IRP_MN_QUERY_REMOVE_DEVICE          =  0x01
IRP_MN_QUERY_RESOURCE_REQUIREMENTS  =  0x0B
IRP_MN_QUERY_RESOURCES              =  0x0A
IRP_MN_QUERY_STOP_DEVICE            =  0x05
IRP_MN_READ_CONFIG                  =  0x0F
IRP_MN_REMOVE_DEVICE                =  0x02
IRP_MN_SCSI_CLASS                   =  0x01
IRP_MN_SET_LOCK                     =  0x12
IRP_MN_START_DEVICE                 =  0x00
IRP_MN_STOP_DEVICE                  =  0x04
IRP_MN_SURPRISE_REMOVAL             =  0x17
IRP_MN_WRITE_CONFIG                 =  0x10

;// POWER minor function codes

IRP_MN_WAIT_WAKE                    =  0x00
IRP_MN_POWER_SEQUENCE               =  0x01
IRP_MN_SET_POWER                    =  0x02
IRP_MN_QUERY_POWER                  =  0x03

;// WMI minor function codes under IRP_MJ_SYSTEM_CONTROL

IRP_MN_QUERY_ALL_DATA               =  0x00
IRP_MN_QUERY_SINGLE_INSTANCE        =  0x01
IRP_MN_CHANGE_SINGLE_INSTANCE       =  0x02
IRP_MN_CHANGE_SINGLE_ITEM           =  0x03
IRP_MN_ENABLE_EVENTS                =  0x04
IRP_MN_DISABLE_EVENTS               =  0x05
IRP_MN_ENABLE_COLLECTION            =  0x06
IRP_MN_DISABLE_COLLECTION           =  0x07
IRP_MN_REGINFO                      =  0x08
IRP_MN_EXECUTE_METHOD               =  0x09
IRP_MN_REGINFO_EX                   =  0x0b
;// WMI minor function
IRP_MN_SET_TRACE_NOTIFY             =  0x0A

;// **********************************************

;//   Type -  

IO_TYPE_ADAPTER                 =  0x00000001
IO_TYPE_CONTROLLER              =  0x00000002
IO_TYPE_DEVICE                  =  0x00000003
IO_TYPE_DRIVER                  =  0x00000004
IO_TYPE_FILE                    =  0x00000005
IO_TYPE_IRP                     =  0x00000006
IO_TYPE_MASTER_ADAPTER          =  0x00000007
IO_TYPE_OPEN_PACKET             =  0x00000008
IO_TYPE_TIMER                   =  0x00000009
IO_TYPE_VPB                     =  0x0000000a
IO_TYPE_ERROR_LOG               =  0x0000000b
IO_TYPE_ERROR_MESSAGE           =  0x0000000c
IO_TYPE_DEVICE_OBJECT_EXTENSION =  0x0000000d

;// Define flags 

IO_FORCE_ACCESS_CHECK           =  0x0001
IO_OPEN_PAGING_FILE             =  0x0002
IO_OPEN_TARGET_DIRECTORY        =  0x0004
IO_NO_PARAMETER_CHECKING        =  0x0100
IO_REPARSE                      =  0x0
IO_REMOUNT                      =  0x1
IO_CHECK_CREATE_PARAMETERS      =  0x0200
IO_ATTACH_DEVICE                =  0x0400
IO_IGNORE_SHARE_ACCESS_CHECK    =  0x0800  ;// Ignores share access checks on opens.
IO_ATTACH_DEVICE_API            =  0x80000000

;// Actions accepted by DRIVER_WRITE_PENDING

IO_DUMP_WRITE_FULFILL     =  0   ;// fulfill IO request as if DRIVER_WAIT
IO_DUMP_WRITE_START       =  1   ;// start new IO
IO_DUMP_WRITE_RESUME      =  2   ;// resume pending IO
IO_DUMP_WRITE_FINISH      =  3   ;// finish pending IO
IO_DUMP_WRITE_INIT        =  4   ;// initialize locals
IO_DUMP_WRITE_DATA_PAGES  =  2


;// Define Driver Object (DRVO) flags

DRVO_UNLOAD_INVOKED             =  0x00000001
DRVO_LEGACY_DRIVER              =  0x00000002
DRVO_BUILTIN_DRIVER             =  0x00000004    ;// Driver objects for Hal, PnP Mgr
DRVO_REINIT_REGISTERED          =  0x00000008
DRVO_INITIALIZED                =  0x00000010
DRVO_BOOTREINIT_REGISTERED      =  0x00000020
DRVO_LEGACY_RESOURCES           =  0x00000040
DRVO_BASE_FILESYSTEM_DRIVER     =  0x00000080   ;// A driver that is at the bottom of the filesystem stack.

;// Device Object (DO) flags

DO_VERIFY_VOLUME                =  0x00000002
DO_BUFFERED_IO                  =  0x00000004
DO_EXCLUSIVE                    =  0x00000008
DO_DIRECT_IO                    =  0x00000010
DO_MAP_IO_BUFFER                =  0x00000020
DO_DEVICE_HAS_NAME              =  0x00000040
DO_DEVICE_INITIALIZING          =  0x00000080
DO_SYSTEM_BOOT_PARTITION        =  0x00000100
DO_LONG_TERM_REQUESTS           =  0x00000200
DO_NEVER_LAST_DEVICE            =  0x00000400
DO_SHUTDOWN_REGISTERED          =  0x00000800
DO_BUS_ENUMERATED_DEVICE        =  0x00001000
DO_POWER_PAGABLE                =  0x00002000
DO_POWER_INRUSH                 =  0x00004000
DO_POWER_NOOP                   =  0x00008000
DO_LOW_PRIORITY_FILESYSTEM      =  0x00010000
DO_XIP                          =  0x00020000

;// Define the Device Object Extension Flags

DOE_UNLOAD_PENDING              =  0x00000001
DOE_DELETE_PENDING              =  0x00000002
DOE_REMOVE_PENDING              =  0x00000004
DOE_REMOVE_PROCESSED            =  0x00000008
DOE_START_PENDING               =  0x00000010
DOE_STARTIO_REQUESTED           =  0x00000020
DOE_STARTIO_REQUESTED_BYKEY     =  0x00000040
DOE_STARTIO_CANCELABLE          =  0x00000080
DOE_STARTIO_DEFERRED            =  0x00000100  ;// Use non-recursive startio
DOE_STARTIO_NO_CANCEL           =  0x00000200  ;// Pass non-cancelable IRP to startio

;// Define the Device Object characteristics flags

FILE_REMOVABLE_MEDIA                     =  0x00000001
FILE_READ_ONLY_DEVICE                    =  0x00000002
FILE_FLOPPY_DISKETTE                     =  0x00000004
FILE_WRITE_ONCE_MEDIA                    =  0x00000008
FILE_REMOTE_DEVICE                       =  0x00000010
FILE_DEVICE_IS_MOUNTED                   =  0x00000020
FILE_VIRTUAL_VOLUME                      =  0x00000040
FILE_AUTOGENERATED_DEVICE_NAME           =  0x00000080
FILE_DEVICE_SECURE_OPEN                  =  0x00000100
FILE_CHARACTERISTIC_PNP_DEVICE           =  0x00000800
FILE_CHARACTERISTIC_TS_DEVICE            =  0x00001000
FILE_CHARACTERISTIC_WEBDAV_DEVICE        =  0x00002000
FILE_CHARACTERISTIC_CSV                  =  0x00010000
FILE_DEVICE_ALLOW_APPCONTAINER_TRAVERSAL =  0x00020000
FILE_PORTABLE_DEVICE                     =  0x00040000

;// Define File Object (FO) flags

FO_FILE_OPEN                    =  0x00000001
FO_SYNCHRONOUS_IO               =  0x00000002
FO_ALERTABLE_IO                 =  0x00000004
FO_NO_INTERMEDIATE_BUFFERING    =  0x00000008
FO_WRITE_THROUGH                =  0x00000010
FO_SEQUENTIAL_ONLY              =  0x00000020
FO_CACHE_SUPPORTED              =  0x00000040
FO_NAMED_PIPE                   =  0x00000080
FO_STREAM_FILE                  =  0x00000100
FO_MAILSLOT                     =  0x00000200
FO_GENERATE_AUDIT_ON_CLOSE      =  0x00000400
FO_DIRECT_DEVICE_OPEN           =  0x00000800
FO_FILE_MODIFIED                =  0x00001000
FO_FILE_SIZE_CHANGED            =  0x00002000
FO_CLEANUP_COMPLETE             =  0x00004000
FO_TEMPORARY_FILE               =  0x00008000
FO_DELETE_ON_CLOSE              =  0x00010000
FO_OPENED_CASE_SENSITIVE        =  0x00020000
FO_HANDLE_CREATED               =  0x00040000
FO_FILE_FAST_IO_READ            =  0x00080000
FO_RANDOM_ACCESS                =  0x00100000
FO_FILE_OPEN_CANCELLED          =  0x00200000
FO_VOLUME_OPEN                  =  0x00400000
FO_FILE_OBJECT_HAS_EXTENSION    =  0x00800000
FO_REMOTE_ORIGIN                =  0x01000000

;// Define I/O Request Packet (IRP) flags

IRP_NOCACHE                     =  0x00000001
IRP_PAGING_IO                   =  0x00000002
IRP_MOUNT_COMPLETION            =  0x00000002
IRP_SYNCHRONOUS_API             =  0x00000004
IRP_ASSOCIATED_IRP              =  0x00000008
IRP_BUFFERED_IO                 =  0x00000010
IRP_DEALLOCATE_BUFFER           =  0x00000020
IRP_INPUT_OPERATION             =  0x00000040
IRP_SYNCHRONOUS_PAGING_IO       =  0x00000040
IRP_CREATE_OPERATION            =  0x00000080
IRP_READ_OPERATION              =  0x00000100
IRP_WRITE_OPERATION             =  0x00000200
IRP_CLOSE_OPERATION             =  0x00000400
IRP_DEFER_IO_COMPLETION         =  0x00000800
IRP_OB_QUERY_NAME               =  0x00001000
IRP_HOLD_DEVICE_QUEUE           =  0x00002000
IRP_RETRY_IO_COMPLETION         =  0x00004000
IRP_HIGH_PRIORITY_PAGING_IO     =  0x00008000
IRP_VERIFIER_MASK               =  0xC0000000
IRP_SET_USER_EVENT              =  IRP_CLOSE_OPERATION

;// Define I/O request packet (IRP) alternate flags for allocation control.

IRP_QUOTA_CHARGED               =  0x01
IRP_ALLOCATED_MUST_SUCCEED      =  0x02
IRP_ALLOCATED_FIXED_SIZE        =  0x04
IRP_LOOKASIDE_ALLOCATION        =  0x08

;// Define stack location control flags

SL_PENDING_RETURNED             =  0x01
SL_ERROR_RETURNED               =  0x02
SL_INVOKE_ON_CANCEL             =  0x20
SL_INVOKE_ON_SUCCESS            =  0x40
SL_INVOKE_ON_ERROR              =  0x80

;// Define flags for various functions

SL_FORCE_ACCESS_CHECK           =  0x01
SL_OPEN_PAGING_FILE             =  0x02
SL_OPEN_TARGET_DIRECTORY        =  0x04
SL_CASE_SENSITIVE               =  0x80
SL_KEY_SPECIFIED                =  0x01
SL_OVERRIDE_VERIFY_VOLUME       =  0x02
SL_WRITE_THROUGH                =  0x04
SL_FT_SEQUENTIAL_WRITE          =  0x08
SL_READ_ACCESS_GRANTED          =  0x01
SL_WRITE_ACCESS_GRANTED         =  0x04    ;// Gap for SL_OVERRIDE_VERIFY_VOLUME
SL_FAIL_IMMEDIATELY             =  0x01
SL_EXCLUSIVE_LOCK               =  0x02
SL_RESTART_SCAN                 =  0x01
SL_RETURN_SINGLE_ENTRY          =  0x02
SL_INDEX_SPECIFIED              =  0x04
SL_WATCH_TREE                   =  0x01
SL_ALLOW_RAW_MOUNT              =  0x01

;//  DeviceType   DEVICE_OBJECT 
;//   IoCreateDevice()

FILE_DEVICE_8042_PORT           =  0x00000027
FILE_DEVICE_ACPI                =  0x00000032
FILE_DEVICE_BATTERY             =  0x00000029
FILE_DEVICE_BEEP                =  0x00000001
FILE_DEVICE_BUS_EXTENDER        =  0x0000002a
FILE_DEVICE_CD_ROM              =  0x00000002
FILE_DEVICE_CD_ROM_FILE_SYSTEM  =  0x00000003
FILE_DEVICE_CHANGER             =  0x00000030
FILE_DEVICE_CONTROLLER          =  0x00000004
FILE_DEVICE_DATALINK            =  0x00000005
FILE_DEVICE_DFS                 =  0x00000006
FILE_DEVICE_DFS_FILE_SYSTEM     =  0x00000035
FILE_DEVICE_DFS_VOLUME          =  0x00000036
FILE_DEVICE_DISK                =  0x00000007
FILE_DEVICE_DISK_FILE_SYSTEM    =  0x00000008
FILE_DEVICE_DVD                 =  0x00000033
FILE_DEVICE_FILE_SYSTEM         =  0x00000009
FILE_DEVICE_FIPS                =  0x0000003a
FILE_DEVICE_FULLSCREEN_VIDEO    =  0x00000034
FILE_DEVICE_INPORT_PORT         =  0x0000000a
FILE_DEVICE_KEYBOARD            =  0x0000000b
FILE_DEVICE_KS                  =  0x0000002f
FILE_DEVICE_KSEC                =  0x00000039
FILE_DEVICE_MAILSLOT            =  0x0000000c
FILE_DEVICE_MASS_STORAGE        =  0x0000002d
FILE_DEVICE_MIDI_IN             =  0x0000000d
FILE_DEVICE_MIDI_OUT            =  0x0000000e
FILE_DEVICE_MODEM               =  0x0000002b
FILE_DEVICE_MOUSE               =  0x0000000f
FILE_DEVICE_MULTI_UNC_PROVIDER  =  0x00000010
FILE_DEVICE_NAMED_PIPE          =  0x00000011
FILE_DEVICE_NETWORK             =  0x00000012
FILE_DEVICE_NETWORK_BROWSER     =  0x00000013
FILE_DEVICE_NETWORK_FILE_SYSTEM =  0x00000014
FILE_DEVICE_NETWORK_REDIRECTOR  =  0x00000028
FILE_DEVICE_NULL                =  0x00000015
FILE_DEVICE_PARALLEL_PORT       =  0x00000016
FILE_DEVICE_PHYSICAL_NETCARD    =  0x00000017
FILE_DEVICE_PRINTER             =  0x00000018
FILE_DEVICE_SCANNER             =  0x00000019
FILE_DEVICE_SCREEN              =  0x0000001c
FILE_DEVICE_SERENUM             =  0x00000037
FILE_DEVICE_SERIAL_MOUSE_PORT   =  0x0000001a
FILE_DEVICE_SERIAL_PORT         =  0x0000001b
FILE_DEVICE_SMARTCARD           =  0x00000031
FILE_DEVICE_SMB                 =  0x0000002e
FILE_DEVICE_SOUND               =  0x0000001d
FILE_DEVICE_STREAMS             =  0x0000001e
FILE_DEVICE_TAPE                =  0x0000001f
FILE_DEVICE_TAPE_FILE_SYSTEM    =  0x00000020
FILE_DEVICE_TERMSRV             =  0x00000038
FILE_DEVICE_TRANSPORT           =  0x00000021
FILE_DEVICE_UNKNOWN             =  0x00000022
FILE_DEVICE_VDM                 =  0x0000002c
FILE_DEVICE_VIDEO               =  0x00000023
FILE_DEVICE_VIRTUAL_DISK        =  0x00000024
FILE_DEVICE_WAVE_IN             =  0x00000025
FILE_DEVICE_WAVE_OUT            =  0x00000026

;// IoCompleteRequest() priority increment

IO_CD_ROM_INCREMENT      =  1
IO_DISK_INCREMENT        =  1
IO_KEYBOARD_INCREMENT    =  6
IO_MAILSLOT_INCREMENT    =  2
IO_MOUSE_INCREMENT       =  6
IO_NAMED_PIPE_INCREMENT  =  2
IO_NETWORK_INCREMENT     =  2
IO_NO_INCREMENT          =  0
IO_PARALLEL_INCREMENT    =  1
IO_SERIAL_INCREMENT      =  2
IO_SOUND_INCREMENT       =  8
IO_VIDEO_INCREMENT       =  1

EVENT_INCREMENT          =  1
SEMAPHORE_INCREMENT      =  1

;// Define interface type

Internal           =  0
Isa                =  1
Eisa               =  2
MicroChannel       =  3
TurboChannel       =  4
PCIBus             =  5
VMEBus             =  6
NuBus              =  7
PCMCIABus          =  8
Cbus               =  9
MPIBus             =  10
MPSABus            =  11
ProcessorInternal  =  12
InternalPowerBus   =  13
PNPISABus          =  14
PNPBus             =  15
Vmcs               =  16

if 0
KIRQL sIrql;
KAFFINITY sAffinity;
ULONG uVector;
ULONG uInterruptNumber = 1;

uVector = HalGetInterruptVector(EisaConfig,0,uInterruptNumber,uInterruptNumber,&sIrql,&sAffinity);
uVector = uVector & 0xFF;
end if

;// Wdf01000!_BUS_DATA_TYPE  HAL.DLL
Cmos              =  0
EisaConfig        =  1
Pos               =  2
CbusConfig        =  3
PCIConfig         =  4
VMEConfig         =  5
NuBusConfig       =  6
PCMCIAConfig      =  7
MPIConfig         =  8
MPSAConfig        =  9
PNPISAConfig      =  10
SgiInternalConfig =  11


;// Define irq priority

IrqPriorityLow     =  1
IrqPriorityNormal  =  2
IrqPriorityHigh    =  3

;// Define IO_ALLOCATION_ACTION
KeepObject                     =  1
DeallocateObject               =  2
DeallocateObjectKeepRegisters  =  3
   

;//*******************************************
;//***************  MACRO  *******************

macro CompleteReq Status, Info {
       mov     rax,[pIrp]
       mov     qword[rax + _IRP.IoStatus],Status
       mov     qword[rax + _IRP.IoStatus+8],Info
       invoke  IoCompleteRequest,[pIrp],IO_NO_INCREMENT }

;//*******************************************

struct IO_STACK_LOCATION
   MajorFunction     db  0
   MinorFunction     db  0
   Flags             db  0
   Control           db  0
   Padding           dd  0
   Parameters        db  32 dup(0)
   DeviceObject      dq  0
   FileObject        dq  0
   CompletionRoutine dq  0
   Context           dq  0 
ends 

struct UNICODE_STRING
   Length           dw  0
   MaximumLength    dw  0
   Reserved         dd  0
   Buffer           dq  0
ends

struct LIST_ENTRY
   Flink      dq  0
   Blink      dq  0 
ends

struct KDPC_DATA
   DpcListHead      LIST_ENTRY
   DpcLock          dq  0
   DpcQueueDepth    dd  0
   DpcCount         dd  0
ends

struct KDPC
   Type             db  0
   Importance       db  0
   Number           dw  0
   Padding          dd  0
   DpcListEntry     LIST_ENTRY
   DeferredRoutine  dq  0 
   DeferredContext  dq  0 
   SystemArgument1  dq  0 
   SystemArgument2  dq  0 
   DpcData          dq  0 
ends

struct KPCR
   GdtBase                 dq  0    ; KGDTENTRY64
   TssBase                 dq  0    ; KTSS64
   UserRsp                 dq  0
   Self                    dq  0    ; KPCR
   CurrentPrcb             dq  0    ; KPRCB
   LockArray               dq  0    ; KSPIN_LOCK_QUEUE
   Used_Self               dq  0
   IdtBase                 dq  0    ; KIDTENTRY64
   Unused                  dq  0,0
   Irql                    db  0
   SecondCacheAssoc        db  0
   ObsoleteNumber          db  0
   Fill0                   db  0
   Unused0                 dd  0,0,0
   MajorVersion            dw  0
   MinorVersion            dw  0
   StallScaleFactor        dd  0
   Unused1                 dq  0,0,0
   KernelReserved          dd  15 dup(0)
   SecondLevelCacheSize    dd  0
   HalReserved             dd  16 dup(0)
   Unused2                 dq  0
   KdVersionBlock          dq  0
   Unused3                 dq  0
   PcrAlign1               dd  24 dup(0)
   Prcb                    dq  0           ; struct KPRCB
ends   

struct DRIVER_EXTENSION
   DriverObject          dq  0  
   AddDevice             dq  0  
   Count                 dq  0  
   ServiceKeyName        UNICODE_STRING
   ClientDriverExtension dq  0    ; IO_CLIENT_EXTENSION
   FsFilterCallbacks     dq  0    ; FS_FILTER_CALLBACKS
ends   
;//----------------------------

struct DRIVER_OBJECT
   Type                  dw  0
   Size                  dw  0
   Padding               dd  0  
   DeviceObject          dq  0
   Flags                 dq  0
   DriverStart           dq  0
   DriverSize            dq  0
   DriverSection         dq  0  ;<-----//   LDR_DATA_TABLE_ENTRY
   DriverExtension       dq  0
   DriverName            UNICODE_STRING
   HardwareDatabase      dq  0
   FastIoDispatch        dq  0
   DriverInit            dq  0
   DriverStartIo         dq  0
   DriverUnload          dq  0
;//-----------------------------   MajorFunction   rq 28
   IRP_MJ_CREATE                   dq  0   ; 0x00
   IRP_MJ_CREATE_NAMED_PIPE        dq  0   ; 0x01
   IRP_MJ_CLOSE                    dq  0   ; 0x02
   IRP_MJ_READ                     dq  0   ; 0x03
   IRP_MJ_WRITE                    dq  0   ; 0x04
   IRP_MJ_QUERY_INFORMATION        dq  0   ; 0x05
   IRP_MJ_SET_INFORMATION          dq  0   ; 0x06
   IRP_MJ_QUERY_EA                 dq  0   ; 0x07
   IRP_MJ_SET_EA                   dq  0   ; 0x08
   IRP_MJ_FLUSH_BUFFERS            dq  0   ; 0x09
   IRP_MJ_QUERY_VOLUME_INFORMATION dq  0   ; 0x0a
   IRP_MJ_SET_VOLUME_INFORMATION   dq  0   ; 0x0b
   IRP_MJ_DIRECTORY_CONTROL        dq  0   ; 0x0c
   IRP_MJ_FILE_SYSTEM_CONTROL      dq  0   ; 0x0d
   IRP_MJ_DEVICE_CONTROL           dq  0   ; 0x0e
   IRP_MJ_INTERNAL_DEVICE_CONTROL  dq  0   ; 0x0f
   IRP_MJ_SHUTDOWN                 dq  0   ; 0x10
   IRP_MJ_LOCK_CONTROL             dq  0   ; 0x11
   IRP_MJ_CLEANUP                  dq  0   ; 0x12
   IRP_MJ_CREATE_MAILSLOT          dq  0   ; 0x13
   IRP_MJ_QUERY_SECURITY           dq  0   ; 0x14
   IRP_MJ_SET_SECURITY             dq  0   ; 0x15
   IRP_MJ_POWER                    dq  0   ; 0x16
   IRP_MJ_SYSTEM_CONTROL           dq  0   ; 0x17
   IRP_MJ_DEVICE_CHANGE            dq  0   ; 0x18
   IRP_MJ_QUERY_QUOTA              dq  0   ; 0x19
   IRP_MJ_SET_QUOTA                dq  0   ; 0x1a
   IRP_MJ_PNP                      dq  0   ; 0x1b
ends

;// https://learn.microsoft.com/en-us/windows-hardware/drivers/ddi/wdm/ns-wdm-_device_object

struct DEVOBJ_EXTENSION
   Type             dw  0
   Size             dw  0
   Reserved         dd  0
   DeviceObject     dq  0
   PowerFlags       dq  0
   Dope             dq  0         ; DEVICE_OBJECT_POWER_EXTENSION
   ExtensionFlags   dq  0
   DeviceNode       dq  0
   AttachedTo       dq  0         ; DEVICE_OBJECT
   StartIoCount     dd  0
   StartIoKey       dd  0
   StartIoFlags     dq  0
   Vpb              dq  0
   DependentList    LIST_ENTRY
   ProviderList     LIST_ENTRY
ends   
   
struct DEVICE_OBJECT
   Type                  dw  0
   Size                  dw  0
   ReferenceCount        dd  0
   DriverObject          dq  0   ; DRIVER_OBJECT
   NextDevice            dq  0   ; DEVICE_OBJECT
   AttachedDevice        dq  0   ; DEVICE_OBJECT
   CurrentIrp            dq  0   ; IRP
   Timer                 dq  0   ; IO_TIMER
   Flags                 dd  0
   Characteristics       dd  0
   Vpb                   dq  0
   DeviceExtension       dq  0   ;<------//       NonPagedPool
   DeviceType            dd  0           ;  ^^^^   fn IoCreateDevice()
   StackSize             dd  0
   Queue                 LIST_ENTRY
   AlignRequirement      dq  0
   DeviceQueue           db  40 dup(0)
   Dpc                   KDPC
   ActiveThreadCount     dq  0
   SecurityDescriptor    dq  0
   DeviceLock            db  24 dup(0)  ; KEVENT
   SectorSize            dq  0
   DeviceObjectExtension dq  0   ; DEVOBJ_EXTENSION
   Reserved              dq  0
ends 

struct IO_STATUS_BLOCK
   Status            dq  0
   Information       dq  0
ends 

struct _IRP
   Type                  dw  0
   Size                  dw  0
   Padding               dd  0   
   MdlAddress            dq  0      ; MDL
   Flags                 dq  0
   AssociatedIrp         dq  0
   ThreadListEntry       LIST_ENTRY
   IoStatus              IO_STATUS_BLOCK
   RequestorMode         db  0
   PendingReturned       db  0   ;<-------
   StackCount            db  0
   CurrentLocation       db  0
   Cancel                db  0
   CancelIrql            db  0
   ApcEnvironment        db  0
   AllocationFlags       db  0
   UserIosb              dq  0      ; IO_STATUS_BLOCK
   UserEvent             dq  0
   Overlay               dq  2 dup(0)
   CancelRoutine         dq  0
   UserBuffer            dq  0
   ;
;   +0x078 Tail.Overlay
   ;
   DriverContext         dq  4 dup(0)
   Thread                dq  0        ; ETHREAD
   AuxiliaryBuffer       dq  0
   ListEntry             LIST_ENTRY
   CurrentStackLocation  dq  0        ; IO_STACK_LOCATION
   OriginalFileObject    dq  0        ; FILE_OBJECT
ends 

struct FILE_OBJECT
   Type               dw  0
   Size               dw  0
   Padding            dd  0
   DeviceObject       dq  0   ; _DEVICE_OBJECT
   Vpb                dq  0   ; _VPB
   FsContext          dq  0 
   FsContext2         dq  0 
   SectionObjPointer  dq  0   ; _SECTION_OBJECT_POINTERS
   PrivateCacheMap    dq  0 
   FinalStatus        dd  0
   RelatedFileObject  dq  0   ; _FILE_OBJECT
   LockOperation      db  0
   DeletePending      db  0
   ReadAccess         db  0
   WriteAccess        db  0
   DeleteAccess       db  0
   SharedRead         db  0
   SharedWrite        db  0
   SharedDelete       db  0
   Flags              dq  0
   FileName           UNICODE_STRING
   CurrentByteOffset  dq  0 
   Waiters            dd  0
   Busy               dd  0
   LastLock           dq  0 
   Lock               db  0x18 dup(0) ; _KEVENT
   Event              db  0x18 dup(0) ; _KEVENT
   CompletionContext  dq  0           ; _IO_COMPLETION_CONTEXT
   IrpListLock        dq  0
   IrpList            LIST_ENTRY
   FileObjectExt      dq  0 
ends

;// IO_STACK_LOCATION Parameters.XX

struct PARAM_CREATE
   SecurityContext    dq  0  ; IO_SECURITY_CONTEXT
   Options            dq  0
   FileAttributes     dw  0
   ShareAccess        dw  0
   Padding            dd  0
   EaLength           dq  0
ends   
struct PARAM_READ_WRITE
   Length             dq  0
   Key                dq  0
   ByteOffset         dq  0
ends
struct PARAM_READ_WRITE_CONFIG
   WhichSpace         dq  0
   Buffer             dq  0
   Offset             dq  0
   Length             dq  0
ends
struct PARAM_IO_CONTROL
   OutputBuffLength   dq  0
   InputBuffLength    dq  0
   IoControlCode      dq  0
   Type3InputBuff     dq  0
ends

;//--------------------------------------

struct DISPATCHER_HEADER
   Type                 db  0
   TimerControlFlags    db  0
   ThreadControlFlags   db  0
   DebugActive          db  0
   SignalState          dd  0
   WaitListHead         LIST_ENTRY
ends

struct KPROCESS
   Header                  DISPATCHER_HEADER
   ProfileListHead         LIST_ENTRY
   DirectoryTableBase      dq  0
   ThreadListHead          LIST_ENTRY
   ProcessLock             dq  0
   Affinity                db  40 dup(0) ; KAFFINITY_EX
   ReadyListHead           LIST_ENTRY
   SwapListEntry           dq  0
   ActiveProcessors        db  40 dup(0) ; KAFFINITY_EX
   ProcessFlags            dd  0
   BasePriority            db  0
   QuantumReset            db  0
   Visited                 db  0
   Unused3                 db  0
   ThreadSeed              dq  0,0
   IdealNode               dq  0
   IdealGlobalNode         dw  0
   Flags                   db  0
   Unused1                 db  0
   Unused2                 dd  0,0
   StackCount              dd  0
   ProcessListEntry        LIST_ENTRY
   CycleTime               dq  0
   KernelTime              dd  0
   UserTime                dd  0
   InstrumentationCallback dq  0
   LdtSystemDescriptor     dq  0,0       ; KGDTENTRY64
   LdtBaseAddress          dq  0
   LdtProcessLock          db  56 dup(0) ; KGUARDED_MUTEX
   LdtFreeSelectorHint     dw  0
   LdtTableLength          dw  0
ends

struct EPROCESS
   Pcb                     KPROCESS
   ProcessLock             dq  0
   CreateTime              dq  0
   ExitTime                dq  0
   RundownProtect          dq  0
   UniqueProcessId         dq  0
   ActiveProcessLinks      LIST_ENTRY   ; offset 0x188
   ProcessQuotaUsage       dq  0,0
   ProcessQuotaPeak        dq  0,0
   CommitCharge            dq  0
   QuotaBlock              dq  0
   CpuQuotaBlock           dq  0
   PeakVirtualSize         dq  0
   VirtualSize             dq  0
   SessionProcessLinks     LIST_ENTRY
   DebugPort               dq  0
   ExceptionPortData       dq  0
   ObjectTable             dq  0
   Token                   dq  0
   WorkingSetPage          dq  0
   AddressCreationLock     dq  0
   RotateInProgress        dq  0         ; ETHREAD
   ForkInProgress          dq  0         ; ETHREAD
   HardwareTrigger         dq  0
   PhysicalVadRoot         dq  0
   CloneRoot               dq  0
   NumberOfPrivatePages    dq  0
   NumberOfLockedPages     dq  0
   Win32Process            dq  0
   Job                     dq  0
   SectionObject           dq  0
   SectionBaseAddress      dq  0
   Cookie                  dd  0
   UmsScheduledThreads     dd  0
   WorkingSetWatch         dq  0
   Win32WindowStation      dq  0
   InheritedFromPid        dq  0
   LdtInformation          dq  0
   Spare                   dq  0
   ConsoleHostProcess      dq  0
   DeviceMap               dq  0
   EtwDataSource           dq  0
   FreeTebHint             dq  0
   FreeUmsTebHint          dq  0
   PageDirectoryPte        dq  0
   Session                 dq  0
   ImageFileName           db  15 dup(0)
   PriorityClass           db  0
   JobLinks                LIST_ENTRY
   LockedPagesList         dq  0
   ThreadListHead          LIST_ENTRY
   SecurityPort            dq  0
   Wow64Process            dq  0
   ActiveThreads           dd  0
   ImagePathHash           dd  0
   DefaultHardError        dd  0
   LastThreadExitStatus    dd  0
   Peb                     dq  0
   PrefetchTrace           dq  0
   ReadOperationCount      dq  0
   WriteOperationCount     dq  0
   OtherOperationCount     dq  0
   ReadTransferCount       dq  0
   WriteTransferCount      dq  0
   OtherTransferCount      dq  0
   CommitChargeLimit       dq  0
   CommitChargePeak        dq  0
   AweInfo                 dq  0
   SeAuditProcessInfo      dq  0
   Vm                      db  136 dup(0)      ; MMSUPPORT
   MmProcessLinks          LIST_ENTRY
   HighestUserAddress      dq  0
   ModifiedPageCount       dd  0
   Flags2                  dd  0
   Flags                   dd  0
   ExitStatus              dd  0
   VadRoot                 db  64 dup(0)       ; MM_AVL_TABLE
ends

struct LDR_DATA_TABLE_ENTRY
   InLoadOrderLinks        LIST_ENTRY
   InMemoryOrderLinks      LIST_ENTRY
   InInitOrderLinks        LIST_ENTRY
   DllBase                 dq  0
   EntryPoint              dq  0
   SizeOfImage             dq  0
   FullDllName             UNICODE_STRING
   BaseDllName             UNICODE_STRING
   Flags                   dd  0
   LoadCount               dw  0
   TlsIndex                dw  0
   SectionPointer          dq  0
   CheckSum                dq  0
   LoadedImports           dq  0
   EntryPointActContext    dq  0
   PatchInformation        dq  0
   ForwarderLinks          LIST_ENTRY
   ServiceTagLinks         LIST_ENTRY
   StaticLinks             LIST_ENTRY
   ContextInformation      dq  0
   OriginalBase            dq  0
   LoadTime                dq  0
ends

struct PEB_LDR_DATA
   Length                  dd  0
   Initialized             dd  0
   SsHandle                dq  0
   InLoadOrderModuleList   LIST_ENTRY
   InMemoryOrderModuleList LIST_ENTRY
   InInitOrderModuleList   LIST_ENTRY
   EntryInProgress         dq  0
   ShutdownInProgress      dq  0
   ShutdownThreadId        dq  0
ends

struct WORK_QUEUE_ITEM
   List             LIST_ENTRY
   WorkerRoutine    dq  0
   Parameter        dq  0
ends
struct IO_WORKITEM
   WorkItem         WORK_QUEUE_ITEM
   Routine          dq  0
   IoObject         dq  0
   Context          dq  0
   Type             dq  0
ends

struct MDL
   Next             dq  0  ; MDL
   Size             dw  0 
   MdlFlags         dw  0  
   Padding          dd  0
   Process          dq  0  ; EPROCESS
   MappedSystemVa   dq  0
   StartVa          dq  0
   ByteCount        dd  0
   ByteOffset       dd  0
ends

;// MmProbeAndLockPages() operation flags
;//  [in,out] PMDL            MemoryDescriptorList,
;//  [in]     KPROCESSOR_MODE AccessMode,
;//  [in]     LOCK_OPERATION  Operation
;//
;// Void MmUnlockPages(
;//  [in, out] PMDL MemoryDescriptorList

IoReadAccess       =  0
IoWriteAccess      =  1
IoModifyAccess     =  2

KernelMode         =  0
UserMode           =  1
MaximumMode        =  2

;// MDL flags
MDL_MAPPED_TO_SYSTEM_VA       =  0x0001
MDL_PAGES_LOCKED              =  0x0002
MDL_SOURCE_IS_NONPAGED_POOL   =  0x0004
MDL_ALLOCATED_FIXED_SIZE      =  0x0008
MDL_PARTIAL                   =  0x0010
MDL_PARTIAL_HAS_BEEN_MAPPED   =  0x0020
MDL_IO_PAGE_READ              =  0x0040
MDL_WRITE_OPERATION           =  0x0080
MDL_LOCKED_PAGE_TABLES        =  0x0100
MDL_PARENT_MAPPED_SYSTEM_VA   =  0x0100
MDL_FREE_EXTRA_PTES           =  0x0200
MDL_DESCRIBES_AWE             =  0x0400
MDL_IO_SPACE                  =  0x0800
MDL_NETWORK_HEADER            =  0x1000
MDL_MAPPING_CAN_FAIL          =  0x2000
MDL_PAGE_CONTENTS_INVARIANT   =  0x4000
MDL_ALLOCATED_MUST_SUCCEED    =  0x4000
MDL_INTERNAL                  =  0x8000
MDL_MAPPING_FLAGS             =  0x0927


















